# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import six


@six.add_metaclass(abc.ABCMeta)
class MonitorBaseDriver(object):
    """
    Abstract class that all monitoring plugins should implement to have a
    unified interface and as many plugins as we want...
    """

    def __init__(self, username, password, endpoint, **kwargs):
        """
        Initializing the driver. Any monitoring system requires the following
        parameters to call it's api. All these parameters can be passed from the
        configuration file in /etc/freezer/dr.conf
        :param username: Username
        :param password: Password
        :param endpoint: API URL
        :param kwargs: any additional parameters can be passed using this config
        option.
        """
        self.username = username
        self.password = password
        self.endpoint = endpoint
        self.kwargs = kwargs

    @abc.abstractmethod
    def get_data(self):
        """
        Gathering metrics data. making the actual api call to
        the monitoring system and get a list of nodes status.
        """
        pass

    @abc.abstractmethod
    def get_metrics(self):
        """
        return list of metrics used to monitor compute nodes. it's Optional
        not all drivers need to implement this method.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def analyze_nodes(self, nodes):
        """
        Process nodes from get_data and return list of down nodes
        :param nodes: dict of metrics of nodes { 'metric1': nodes,
        'metric2': nodes}
        :return: a list of down nodes
        """
        pass

    @abc.abstractmethod
    def process_failed(self, nodes=[], wait=0):
        """
        Double check the failed nodes again to make sure that nodes are down.
        return a list of down nodes to be passed to the evacuation tool to
        process failed hosts.
        :param nodes: a list contains pre-checked nodes to re-check them again
        :param wait: a configurable a mount of time to wait before doing this
        check to give a chance for the host to recover if there was a minor
        issue.
        :return: a list of nodes to be evacuated, the list will be passed
        directly to the evacuation tool to process them
        """
        pass

    @abc.abstractmethod
    def is_alive(self):
        """
        Plugin should provide a way to make sure that the monitoring system is
        a live or not. It's optional not all drivers need to implement it.
        :return: True or False
        """
        raise NotImplementedError()

    def get_info(self):
        """
        Get Driver information ..
        :return: dict of name, version, author, ...
        """



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from freezer_dr.fencers.common.driver import FencerBaseDriver
from freezer_dr.fencers.drivers.ipmi.ipmitool import IpmiInterface
from oslo_log import log
from oslo_config import cfg

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class IpmiDriver(FencerBaseDriver):
    def __init__(self, node, **kwargs):
        super(IpmiDriver, self).__init__(node, **kwargs)
        self.interface = IpmiInterface(
            node.get('fencer-ip'),
            node.get('fencer-user'),
            node.get('fencer-password'),
            verbose=CONF.debug)

    def force_shutdown(self):
        try:
            self.interface.power_down()
        except Exception as e:
            LOG.error(e)

    def graceful_shutdown(self):
        try:
            self.interface.power_soft()
        except Exception as e:
            LOG.error(e)

    def status(self):
        return self.interface.get_power_status()

    # @todo remove this fn as it's for testing purposes only :)
    def power_on(self):
        self.interface.power_on()

    def get_info(self):
        return {
            'name': 'IPMI Interface driver',
            'version': 1.0,
            'author': 'Hewlett-Packard Development Company, L.P'
        }





# (c) Copyright 2016 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abstract fencer"""

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class FencerBaseDriver(object):

    """Abstract class that all fencer plugins.

    Should be implemented to have a unified interface and as many plugins as
    needed.
    """

    def __init__(self, node, **kwargs):
        """Initialize the driver.

        Any fencer driver requires the following parameters to do the api
        calls. All these parameters can be passed from the configuration
        file in /etc/freezer/dr.conf (default).

        :param node: dict with all node details. (/etc/freezer/servers.yml) ?
        :param kwargs: any additional parameters can be passed using this
                       config option.
        """
        self.node = node
        self.kwargs = kwargs

    @abc.abstractmethod
    def graceful_shutdown(self):
        """Gracefully shutdown the compute node to evacuate it."""

    @abc.abstractmethod
    def force_shutdown(self):
        """Force shutdown the compute node to evacuate it.

        May be you can try force shutdown if the graceful shutdown failed.
        """

    @abc.abstractmethod
    def status(self):
        """Get compute node status.

        Should return 1 if on and 0 if off or -1 if error or unknown power
        status.
        """

    @abc.abstractmethod
    def get_info(self):
        """Get Driver information.

        :return: dict of name, version, author, ...
        """



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from oslo_config import cfg
from oslo_log import log
from freezer_dr.evacuators.common.driver import EvacuatorBaseDriver
from freezer_dr.common.utils import get_os_client
CONF = cfg.CONF
LOG = log.getLogger(__name__)


class StandardEvacuator(EvacuatorBaseDriver):

    def __init__(self, wait, retires, shared_storage, **kwargs):
        super(StandardEvacuator, self).__init__(wait, retires, shared_storage,
                                                **kwargs)
        self.client = get_os_client()

    def get_node_instances(self, node):
        return self.client.get_hypervisor_instances(node)

    def disable_node(self, node):
        return self.client.disable_node(node)

    def get_node_status(self, node):
        return self.client.get_node_status(node)

    def is_node_disabled(self, node):
        return self.client.get_node_status(node)

    def evacuate_nodes(self, nodes):
        return self.client.evacuate(nodes, shared_storage=self.shared_storage)






# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils
from freezer_dr.fencers.common.manager import FencerManager
from time import sleep
from freezer_dr.evacuators.common.utils import get_nodes_details

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class EvacuationManager(object):

    def __init__(self, enable_fencing=True):
        evcuation_conf = CONF.get('evacuation')
        self.driver = importutils.import_object(
            evcuation_conf.get('driver'),
            evcuation_conf.get('wait'),
            evcuation_conf.get('retries'),
            evcuation_conf.get('shared_storage'),
            **evcuation_conf.get('options')
        )
        self.enable_fencing = enable_fencing
        self.wait = evcuation_conf.get('wait')
        self.retires = evcuation_conf.get('retries', 1)
        if self.retires <= 0:
            self.retires = 1

    def evacuate(self, nodes):
        # try to disable node
        # @todo needs more error handling like if the status didn't update or
        # we are unable to disable the node ???
        failed_nodes = []  # maintain nodes that are going to fail at any state
        succeeded_nodes = []
        for node in nodes:
            for i in range(0, self.retires):
                status = self._disable_node(node)
                # if True ( node disabled ) break the loop
                if status:
                    break
                else:
                    status = False
            node['status'] = status
            # make sure the disable request was successful
            if not self.driver.get_node_status(node):
                failed_nodes.append(node)
                nodes.remove(node)  # if the node failed at any step no reason
                # to move it to the next step
            else:
                succeeded_nodes.append(node)

        nodes = succeeded_nodes
        if self.enable_fencing:
            fencer = FencerManager(nodes)
            nodes = fencer.fence()
        """
        @todo this code needs to be commented for the time being till we fix
         nova bug found in state, which always go up afer enable or disable. We
         will use get_node_details for the time being from the main script to
         get nodes details before evacuating ...
        succeeded_nodes = []
        for node in nodes:
            node['instances'] = self.driver.get_node_instances(node)
            succeeded_nodes.append(node)

        nodes = succeeded_nodes
        """
        # Start evacuation calls ...

        for i in range(0, 10):
            try:
                sleep(30)
                evacuated_nodes = self.driver.evacuate_nodes(nodes)
                print "Try Number: ", i
                print evacuated_nodes
            except Exception as e:
                LOG.error(e)
        return evacuated_nodes

    def get_nodes_details(self, nodes):
        """
        To be re-structured after fixing the nova bug !
        :param nodes: list of nodes
        :return: list of node with more details
        """
        return get_nodes_details(nodes)

    def _disable_node(self, node):
        if not self.driver.is_node_disabled(node):
                return self.driver.disable_node(node)
        else:
            True



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils
from freezer_dr.common.yaml_parser import YamlParser
from time import sleep

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class FencerManager(object):

    def __init__(self, nodes):
        self.fencer = CONF.get('fencer')
        self.nodes = nodes
        self.parser = YamlParser(self.fencer.get('credentials_file'))

    def fence(self):
        """
        Try to shutdown nodes and wait for configurable amount of times
        :return: list of nodes and either they are shutdown or failed
        """
        processed_nodes = []
        for node in self.nodes:
            node_details = self.parser.find_server_by_ip(node.get('ip')) or\
                           self.parser.find_server_by_hostname(node.get('host'))
            driver = importutils.import_object(
                self.fencer.get('driver'),
                node=node_details,
                **self.fencer.get('options')
            )
            node['status'] = self.do_shutdown_procedure(driver)
            processed_nodes.append(node)
        return processed_nodes

    def do_shutdown_procedure(self, driver):
        for retry in range(0, self.fencer.get('retries', 1)):
            if driver.status():
                try:
                    driver.graceful_shutdown()
                except Exception as e:
                    LOG.error(e)
            else:
                return True
            # try to wait a pre-configured amount of time before redoing
            # the fence call again :)
            sleep(self.fencer.get('hold_period', 10))
            LOG.info('wait for %d seconds before retrying to gracefully '
                     'shutdown' % self.fencer.get('hold_period', 10))
            LOG.info('Retrying to gracefully shutdown the node.')

        try:
            driver.force_shutdown()
        except Exception as e:
            LOG.error(e)

        if not driver.status():
            return True

        return False




# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from freezer_dr.common import config
from oslo_config import cfg
from oslo_log import log
from freezer_dr.monitors.common.manager import MonitorManager
from freezer_dr.evacuators.common.manager import EvacuationManager
from freezer_dr.notifiers.common.manager import NotificationManager

CONF = cfg.CONF
LOG = log.getLogger(__name__)


def main():
    config.configure()
    config.setup_logging()
    LOG.info('Starting Freezer DR ... ')
    # load and initialize the monitoring driver
    monitor = MonitorManager()
    # Do the monitoring procedure
    # Monitor, analyse, nodes down ?, wait, double check ? evacuate ..
    nodes = monitor.monitor()

    if nodes:
        # @todo put node in maintenance mode :) Not working with virtual
        # deployments
        # Load Fence driver
        # Shutdown the node
        evac = EvacuationManager()
        notify_nodes = evac.get_nodes_details(nodes)
        evac.evacuate(nodes)
        notifier = NotificationManager()
        notifier.notify(notify_nodes, 'success')
    else:
        print "No nodes reported to be down"



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from freezer_dr.fencers.common.driver import FencerBaseDriver
from oslo_log import log
from oslo_config import cfg
import libvirt

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class LibvirtDriver(FencerBaseDriver):
    def __init__(self, node, **kwargs):
        super(LibvirtDriver, self).__init__(node, **kwargs)
        conn_name = kwargs.get('name', None)
        self.connection = libvirt.open(name=conn_name)

    def force_shutdown(self):
        target = self.connection.lookupByName(name=self.node.get('domain-name'))
        return target.destroy()

    def graceful_shutdown(self):
        target = self.connection.lookupByName(name=self.node.get('domain-name'))
        return target.shutdown()

    def status(self):
        target = self.connection.lookupByName(name=self.node.get('domain-name'))
        return target.isActive()

    def get_info(self):
        return {
            'name': 'Libvirt Interface driver',
            'version': 1.0,
            'author': 'Hewlett-Packard Development Company, L.P'
        }





# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import six


@six.add_metaclass(abc.ABCMeta)
class EvacuatorBaseDriver(object):
    """
    Abstract class for all evacuation drivers should implement to have
    a unified interface
    """

    def __init__(self, wait, retries, shared_storage, **kwargs):
        """
        Initialize Evacuation driver with the config args
        :param wait: time in seconds that the evcauator should wait before
        retrying to disable the node
        :param retries: Number of times the evacuator will try to disable the
        compute node
        :param shared_storage: Boolean; True if the compute nodes are running
        on shared storage and False otherwise
        :param kwargs: Dict of arguments that any future driver may need to
        load it from the config file
        :return: None
        """
        self.wait = wait
        self.retries = retries
        self.shared_storage = shared_storage
        self.options = kwargs

    @abc.abstractmethod
    def disable_node(self, node):
        """
        Disable the compute node from accepting any new VMs or requests
        :param node: dict contains node's hostname
        :return: True pr False
        """
        pass

    @abc.abstractmethod
    def is_node_disabled(self, node):
        """
        Check if node is already disabled or not
        :param node: dict contains node's hostname
        :return: True or False
        """
        pass

    @abc.abstractmethod
    def evacuate_nodes(self, nodes):
        """
        Will evacuate all running VMs on the required nodes
        :param nodes: list of nodes
        :return: list of nodes with updated status
        """
        pass

    @abc.abstractmethod
    def get_node_instances(self, node):
        """
        List instances on a compute host
        :param node: dict contains node's hostname
        :return: List contains running VMs on a given node
        """
        pass

    def get_info(self):
        """
        Get Driver Information
        :return: Dict contains driver information
        """
        pass

    @abc.abstractmethod
    def get_node_status(self, node):
        """
        Check the node status and report it
        :param node: dict contains node's hostname
        :return: dict with key 'status': 'True or False'
        """
        pass
